//***********************************************************
//*
//* File:           Friends.java
//* Author:         Abhinav Kamra
//* Contact:        kamra@cs.columbia.edu
//* Update:         10/22/2004
//*
//* Description:    Game model for Project 3, CS4444
//*                 Fall 2004
//*
//*
//***********************************************************


package Friends;

import ui.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.text.NumberFormat;

public final class Friends implements IFCConstants, IFCModel, MouseListener {

    private Class[]                 _classlist;
    private Class[]                 _playerlist;
    private int                     _currRound;
    private int                     _maxrounds;
    private int                     _numplayers;
    private int                     _state;
    private transient IFCUI         _ui;
    private IFCConfiguration        _config;
    private static Random           _random;
    private JTextField              _input;
    private ControlPanel            _control;
    private ViewPanel               _view;
    private boolean                 _registered;
    private static final int        _CMOVING                = 0;
    private static final int        _CWAITING               = 1;
    private static final int        _CFINISHED              = 2;
    private static final int[]      _CSTATES                = { _CMOVING, _CWAITING, _CFINISHED };
    private static final String     _CNAME                  = "Let's Get Together !";
    private static final String     _CPROPERTIES_FILE       = "gamemodel.properties";
    private static final int        _CMIN_ROUNDS            = 1;
    private static final int        _CMAX_ROUNDS            = 500000;
    private static final int        _CMIN_PLAYERS           = 1;
    private static final int        _CMAX_PLAYERS           = 1000;
    private static final int        _CMIN_L		    = 1;
    private static final int        _CMAX_L		    = 500;
    private static final int        _CMIN_W		    = 1;
    private static final int        _CMAX_W		    = 500;

    class Cell
    {
	    boolean[] memorylist;
	    ArrayList players;
	    Cell(int N)
	    {
		    memorylist = new boolean[N];
		    players = new ArrayList();
	    }
    }

    private int L;	// Grid Width
    private int W;	// Grid Height
    private Cell[][] cells;	// Grid
    private PlayerWrapper[] _players;	// The Players

    private int[][] map;
    int mouseX, mouseY;
    int InteractivePlayer;
    Move _tmpmove;
    double cellsize;

    //********************************************
    //*
    //* Constructors
    //*
    //********************************************
    public Friends() throws Exception {
	create(createDefaultConfiguration());
    }

    public Friends(IFCConfiguration __config) throws Exception {
        create(__config);
    }

    public Friends(IFCTournament __tournament) throws Exception {
        run(__tournament);
    }
    
    //********************************************
    //*
    //* Constructor Delegates
    //*
    //********************************************
    public void run(IFCTournament __tournament) throws Exception {

	    /*
	    IFCGameRecord[] games;
	    double starttime;
	
	    _random = new Random();
	    
	    _maxrounds = _config.numRounds();
	    init_energy= _config.InitEnergy();
	    X = _config.GridX();
	    Y = _config.GridY();
	    M = _config.M();
	    K = _config.K();
	    v = _config.v();
	    u = _config.u();
	    p = _config.p();
	    q = _config.q();
	    s = 1;

	    cells = new Cell[X][];
	    for(int i=0;i < X;i++)
	    {
		    cells[i] = new Cell[Y];
		    for(int j=0;j < Y;j++)
			    cells[i][j] = new Cell();
	    }


	    games = __tournament.games();
	    if (games == null) {
		    throw new Exception("Error:  Null game record list");
	    }

	    starttime = System.currentTimeMillis()/60000.0;
	    for(int game=0;game < games.length;game++)
	    {
    	System.err.println("[Game "+ game +" of " + games.length + "]: Starting at " + (System.currentTimeMillis()/60000.0 - starttime) + " minutes from start");
	    population = 0;
		    _playerlist = games[game].players();
		    _numplayers = _playerlist.length;
		    if(X*Y < _numplayers)
			    throw new Exception("More players than Space on Grid");
		    for(int i=0;i < X;i++)
		    {
			    for(int j=0;j < Y;j++)
			    {
				    cells[i][j].foodvalue = 0;
				    cells[i][j].pw = null;
			    }
		    }
		    OrigPlayers = new PlayerEntry[_numplayers];
		    for(int i=0;i < _numplayers;i++)
		    {
			    int x, y;
			    while(true)
			    {
				    x = _random.nextInt(X);
				    y = _random.nextInt(Y);
				    if(cells[x][y].pw == null)
					    break;
			    }
			    population++;
			    cells[x][y].pw = new PlayerWrapper(_playerlist[i]);
			    cells[x][y].pw.register(this, -1);
			    cells[x][y].foodvalue = 0;;
			    cells[x][y].playertype = i;;
			    OrigPlayers[i] = new PlayerEntry(_playerlist[i], cells[x][y].pw.name(), cells[x][y].pw.color());
			    ChangeEnergy(x, y, init_energy);
		    }

		    _currRound = 0;
		    _state = _CMOVING;
		    
		    while (step()) { }
		    games[game].setScores(finalScores());
	    }
    */
        
    } // end -- run

    void createMap(int N)
    {
	    map = new int[N][N];

	    for(int i=0;i < N;i++)
	    {
		    for(int j=0;j < N;j++)
			    map[i][j] = j;
		    map[i][i] = 0;
		    map[i][0] = i;
	    }

	    for(int i=0;i < N;i++)
	    {
		    for(int num=0;num < 2*N;num++)
		    {
			    int x = _random.nextInt(N);
			    int y = _random.nextInt(N);

			    if(x == i || y == i)
				    continue;

			    //Swap
			    int tmp = map[i][x];
			    map[i][x] = map[i][y];
			    map[i][y] = tmp;
		    }
	    }
    }

    void create(IFCConfiguration __config) throws Exception
    {
	    _random = new Random();
	    _config = __config;

	    
	    _maxrounds = _config.maxRounds();
	    L = _config.GridL();
	    W = _config.GridW();

	    _classlist = _config.classList();
	    _playerlist = _config.playerList();
	    _numplayers = _playerlist.length;
	    _tmpmove = new Move(_numplayers);

	    createMap(_numplayers);
	    
	    cells = new Cell[L][];
	    for(int i=0;i < L;i++)
	    {
		    cells[i] = new Cell[W];
		    for(int j=0;j < W;j++)
			    cells[i][j] = new Cell(_numplayers);
	    }
	    
	    _players = new PlayerWrapper[_numplayers];
	    for(int i=0;i < _numplayers;i++)
	    {
		    int x = _random.nextInt(L);
		    int y = _random.nextInt(W);

		    while(cells[x][y].players.size() > 0)
		    {
		    	x = _random.nextInt(L);
		    	y = _random.nextInt(W);
		    }

		    _players[i] = new PlayerWrapper(_playerlist[i]);
		    _players[i].register(this);

		    cells[x][y].players.add(new Integer(i));
		    _players[i].setLocation(x, y);
		    _players[i].SetInfo(i);
	    }

	    _currRound = 0;
	    _state = _CMOVING;
	    postProcessGrid();
	    //_history = new ArrayList();
        
        _control = new ControlPanel();
        _view = new ViewPanel();
	_view.addMouseListener(this);
    }

	public void mouseClicked(MouseEvent e)
	{
		int x = e.getX();
		int y = e.getY();
		x -= _view._MARGIN;
		y -= _view._MARGIN;

		if(x > 0 && x < cellsize*L && y > 0 && y < cellsize*W)
		{
			mouseX = (int)Math.round(Math.floor((double)x / cellsize));
			mouseY = (int)Math.round(Math.floor((double)y / cellsize));
		}
		_view.repaint();
	}

	public void mouseEntered(MouseEvent e)
	{
	}

	public void mouseExited(MouseEvent e)
	{
	}

	public void mousePressed(MouseEvent e)
	{
	}

	public void mouseReleased(MouseEvent e)
	{
	}


    //********************************************
    //*
    //* Initial Configuration
    //*
    //********************************************
    public IFCConfiguration createDefaultConfiguration() throws Exception {
        IFCConfiguration RET = new Configuration();
        String[] toks;
        Class[] classes;
        Class[] players;
        int _MAX;
        Properties properties;
        Random random = new Random();
        ParseValue pv;

        RET.setMaxRoundsBounds(_CMIN_ROUNDS, _CMAX_ROUNDS);
        RET.setNumPlayersBounds(_CMIN_PLAYERS, _CMAX_PLAYERS);

        properties = Util.gatherProperties(_CPROPERTIES_FILE);
        RET.setLogFile(properties.getProperty("LOG_FILE").trim());

        pv = ParseValue.parseIntegerValue(properties.getProperty("MAX_ROUNDS").trim(), _CMIN_ROUNDS, _CMAX_ROUNDS);
        if (!pv.isValid()) {
            throw new Exception("Properties parameter out of range, Number of Rounds");
        }
        RET.setMaxRounds(((Integer) pv.value()).intValue());

        pv = ParseValue.parseIntegerValue(properties.getProperty("L").trim(), _CMIN_L, _CMAX_L);
        if (!pv.isValid()) {
            throw new Exception("Properties parameter out of range, L");
        }
        RET.setGridL(((Integer) pv.value()).intValue());

        pv = ParseValue.parseIntegerValue(properties.getProperty("W").trim(), _CMIN_W, _CMAX_W);
        if (!pv.isValid()) {
            throw new Exception("Properties parameter out of range, W");
        }
        RET.setGridW(((Integer) pv.value()).intValue());

        toks = Util.split(",\t\n ", properties.getProperty("CLASS_LIST").trim());
        _MAX = toks.length;
        classes = new Class[_MAX];
        for (int i=0; i < _MAX; i++) {
            classes[i] = Class.forName(toks[i]);
        }
        RET.setClassList(classes);

        toks = Util.split(",\t\n ", properties.getProperty("PLAYER_LIST").trim());
        _MAX = toks.length;
        if (_MAX < _CMIN_PLAYERS || _MAX > _CMAX_PLAYERS) {
            throw new Exception("Properties parameter out of range, Number of Players (Player List)");
        }
        players = new Class[_MAX];
        for (int i=0; i < _MAX; i++) {
            players[i] = Class.forName(toks[i]);
        }
        RET.setPlayerList(players);

	int[] playercopies;
	playercopies = new int[classes.length];
	for(int i=0;i < players.length;i++)
	{
		for(int j=0;j < classes.length;j++)
		{
			if(players[i] == classes[j])
				playercopies[j]++;
		}
	}
	RET.setPlayerCopies(playercopies);

        return RET;
    }

    //********************************************
    //*
    //* Exposed Methods
    //*
    //********************************************

    public int numPlayers()
    {
	    return _numplayers;
    }

    public int currRound()
    {
	    return _currRound;
    }

    public int maxRounds()
    {
	    return _maxrounds;
    }

    public int L()
    {
	    return L;
    }

    public int W()
    {
	    return W;
    }

    private int indexOf(IFCPlayer __player) throws Exception {
        int _MAX = numPlayers();

        for (int i=0; i < _MAX; i++) {
            if (_players[i].player() == __player) {
                return i;
            }
        }
        throw new Exception("Player not found: "+__player);
    }

    public boolean[] MyInfo(IFCPlayer __player) throws Exception
    {
	    int playerindex = indexOf(__player);
	    boolean[] realinfo = _players[playerindex].GetInfo();
	    boolean[] RET = new boolean[realinfo.length];

	    for(int i=0;i < realinfo.length;i++)
	    {
		    int virtual = map[playerindex][i];
		    RET[virtual] = realinfo[i];
	    }
	    return RET;
    }

    public Random rand()
    {
	    return _random;
    }

    public void print(String __str) throws Exception {
        if (_registered) {
              _ui.print(__str);
        }
    }

    public void println(String __str) throws Exception {
        if (_registered) {
              _ui.println(__str);
        }
    }

    public void println() throws Exception {
        if (_registered) {
              _ui.println();
        }
    }

    //********************************************
    //*
    //* IFCModel
    //*
    //********************************************
    public void register(IFCUI __ui) throws Exception {
        _ui = __ui;
        _ui.register(this);
        _registered = true;

//        create(createDefaultConfiguration());

        println("[Player Configuration]: ");
        for (int i=0; i < _numplayers; i++) {
            println("\t[Player" + i + "]: " + _players[i].name());
        }
        refresh();
    }

    public String name() throws Exception {
        return _CNAME;
    }

    public JPanel exportControlPanel() throws Exception {
        return _control;
    }

    public JPanel exportViewPanel() throws Exception {
        return _view;
    }

    public JButton[] exportTools() throws Exception {
        return _control.exportTools();
    }
    
    public JMenu exportMenu() throws Exception {
        return null;
    }

    public IFCConfiguration exportConfiguration() throws Exception {
        return _config;
    }
    
    
    private void refresh() throws Exception {
        if (_registered) {
            _ui.refresh();
        }
    } 
    //********************************************
    //*
    //* Private Methods
    //*
    //********************************************


    private void reset() throws Exception {
        if (_registered) {
            _ui.reset();
        }
    }

    boolean WithProb(double p)
    {
	    double tmp = _random.nextDouble();
	    return ((tmp <= p) ? true : false );
    }

    int min(int a, int b)
    {
	    return (a < b)?a:b;
    }

    double min(double a, double b)
    {
	    return (a < b)?a:b;
    }

    int max(int a, int b)
    {
	    return (a > b)?a:b;
    }

    boolean[] UnionInfo(boolean[] a, boolean[] b) throws Exception
    {
	    for(int i=0;i < a.length;i++)
	    {
		    a[i] |= b[i];
	    }
	    return a;
    }

    void postProcessGrid() throws Exception
    {
	    for(int i=0;i < L;i++)
	    {
	    	for(int j=0;j < W;j++)
		{
			boolean[] allinfo = new boolean[_numplayers];
			int size = cells[i][j].players.size();
			
			for(int n=0;n < size;n++)
			{
				allinfo = UnionInfo(allinfo, _players[((Integer)cells[i][j].players.get(n)).intValue()].GetInfo());
			}

			cells[i][j].memorylist = UnionInfo(cells[i][j].memorylist, allinfo);

			if(size > 1)
			{
				for(int n=0;n < size;n++)
				{
					if(_players[((Integer)cells[i][j].players.get(n)).intValue()].SetInfo(allinfo) == _numplayers)
						_state = _CFINISHED;
				}
			}
		}
	    }
    }

    void PrepareMove(int player, int x, int y)
    {
	    for(int i=0;i < _CDIRECTIONS.length;i++)
	    {
		    int currx = (x + _CXTrans[i] + L)%L;
		    int curry = (y + _CYTrans[i] + W)%W;

		    for(int j=0;j < _numplayers;j++)
			    _tmpmove.information[i][map[player][j]] = cells[currx][curry].memorylist[j];

		    _tmpmove.friends[i].clear();
		    for(int j=0;j < cells[currx][curry].players.size();j++)
		    {
			    int friendnumber =((Integer)cells[currx][curry].players.get(j)).intValue();
			    _tmpmove.friends[i].add(new Integer(map[player][friendnumber]));
		    }
	    }
    }

    void processMove(int player, int movetype) throws Exception
    {
	    int currx = _players[player].x();
	    int curry = _players[player].y();

	    if(movetype < _CSTAYPUT || movetype > _CSOUTHEAST)
		    movetype = _CSTAYPUT;

	    int newx = (currx + _CXTrans[movetype] + L)%L;
	    int newy = (curry + _CYTrans[movetype] + W)%W;

	    _players[player].setLocation(newx, newy);

	    int index = cells[currx][curry].players.indexOf(new Integer(player));

	    cells[newx][newy].players.add(cells[currx][curry].players.remove(index));
    }

    //********************************************
    //*
    //* State Transition
    //*
    //********************************************

    private boolean step() throws Exception {

        StringBuffer SB = new StringBuffer();

        switch (_state) {
            case _CWAITING: {
                println("Please Make A Move, ");
                return false;
            }

            case _CMOVING:  {
//				    println("------ Round " + _currRound + " ------");
				    int[] results = new int[_numplayers];
				    for(int i=0;i < _numplayers;i++)
				    {
					    PrepareMove(i, _players[i].x(), _players[i].y());
					    results[i] = _players[i].move(_tmpmove);
				    }

				    for(int i=0;i < _numplayers;i++)
				    {
					    processMove(i, results[i]);
				    }
				    postProcessGrid();
				    /*
				    SB = new StringBuffer();
				    SB.append("Player " + cells[x][y].playertype + " at[" + (x+1) + ", " + (y+1));
				    SB.append("]: ");
				    if(_move != null)
				    SB.append(_move.toString());
				    println(new String(SB));
				    */
				    _currRound++;
				    if(_currRound >= maxRounds())
					    _state = _CFINISHED;
				    break;
			    }
	}
	return (_state != _CFINISHED);

    }
    

    //********************************************
    //*
    //* View Panel
    //*
    //********************************************
    private final class ViewPanel extends JPanel implements Serializable {
        final int           _CWIDTH                         = 600;
        final int           _CHEIGHT                        = 600;
        final int           _CYMARGIN                       = 200;
	public final int	    _MARGIN			    = 20;
        final Font          _CVIEW_FONT                     = new Font("Courier", Font.BOLD, 5);
        final Font          _CAMOEBA_FONT                   = new Font("Courier", Font.BOLD, (max(L, W) >= 50) ? (max(L, W) >= 60) ? 8 : 9 : 10);
        final int           _CHOFFSET                       = _CAMOEBA_FONT.getSize() / 3;
        final int           _CVOFFSET                       = _CAMOEBA_FONT.getSize();
        final Color         _CBLACK                         = new Color(0.0f, 0.0f, 0.0f);
        final Color         _CBACKGROUND_COLOR              = new Color(1.0f, 1.0f, 1.0f);
        final int           _COUTLINE_THICKNESS             = 0;

        //********************************************
        //*
        //* Constructor
        //*
        //********************************************            
        public ViewPanel() throws Exception {
            super();                                        
            setLayout(new BorderLayout());
            setBackground(_CBACKGROUND_COLOR);
            setBorder(BorderFactory.createCompoundBorder(BorderFactory.createRaisedBevelBorder(),
                                                         BorderFactory.createLoweredBevelBorder()));
            setPreferredSize(new Dimension(_CWIDTH, _CHEIGHT));
            setMinimumSize(new Dimension(_CWIDTH, _CHEIGHT));
            setFont(_CAMOEBA_FONT);
        }       
        
        //********************************************
        //*
        //* paint() Override
        //*
        //********************************************            
        public void paintComponent(Graphics __g)
	{
            try {
                super.paintComponent(__g);
                
                int width = getWidth() - _MARGIN;
                int height = getHeight() - _MARGIN - _CYMARGIN;
//                double _ratio = (width < height ? (double) width / (double) max(L, W) : (double) height / (double) max(L, W));
		double _ratio = min( (double)width / (double)L, (double)height / (double)W );
		cellsize = _ratio;
                
		__g.drawLine(_MARGIN, _MARGIN,_MARGIN, _MARGIN + (int) (W * _ratio));
		__g.drawLine(_MARGIN, _MARGIN, _MARGIN + (int) (L * _ratio), _MARGIN);

		for(int x=0;x < L;x++)
		{
			__g.setColor(_CBLACK);
			__g.drawLine(_MARGIN + (int) ((x+1) * _ratio), _MARGIN,_MARGIN + (int) ((x+1) * _ratio), _MARGIN + (int) (W * _ratio));
			__g.drawString(Integer.toString(x), _MARGIN + (int) (x * _ratio) + _CHOFFSET, _CVOFFSET);
			for(int y=0;y < W;y++)
			{
				if(x == 0)
				{
					__g.setColor(_CBLACK);
					__g.drawLine(_MARGIN, _MARGIN + (int) ((y+1) * _ratio), _MARGIN + (int) (L * _ratio), _MARGIN + (int) ((y+1) * _ratio));
					__g.drawString(Integer.toString(y), _CHOFFSET, _MARGIN + (int) (y * _ratio) + _CVOFFSET);
				}
				if(cells[x][y].players.size() > 0)
				{
					//__g.setColor(_CBLACK);
					//__g.fillRect(_MARGIN + (int) (x * _ratio), _MARGIN + (int) (y * _ratio), (int) _ratio, (int) _ratio);

					int player = ((Integer)cells[x][y].players.get(0)).intValue();
					__g.setColor(_players[player].color());
					__g.fillRect(_MARGIN + (int) (x * _ratio) + _COUTLINE_THICKNESS, _MARGIN + (int) (y * _ratio) + _COUTLINE_THICKNESS, (int) _ratio - _COUTLINE_THICKNESS * 2, (int) _ratio - _COUTLINE_THICKNESS * 2);
					__g.setColor(_CBLACK);
					__g.drawString(Integer.toString(player), _MARGIN + (int) (x * _ratio) + _CHOFFSET, _MARGIN + (int) (y * _ratio) + _CVOFFSET);
				}
			}
		}

		int h = (_CYMARGIN - _MARGIN) / _numplayers;
		int w = (getWidth()/2) / _numplayers;
		int x0 = getWidth() / 8;
		int y0 = getHeight() - _CYMARGIN + _MARGIN/2;
		for(int i=0;i < _numplayers;i++)
		{
			boolean[] info = _players[i].GetInfo();
			__g.setColor(_players[i].color());
			for(int j=0;j < _numplayers;j++)
			{
				if(info[j])
				{
					__g.fillRect(x0 + j*w, y0 + i*h, w, h);
				}
			}
			__g.drawString("Player " + Integer.toString(i), x0 - 3*_MARGIN, y0 + (i+1)*h);
		}
		__g.setColor(_CBLACK);
		__g.drawLine(x0 + _numplayers*w + _MARGIN, y0, x0 + _numplayers*w + _MARGIN, getHeight());
		
		// Display the memory info for a cell from some user's perspective
		int x1 = getWidth() - w;
		int y1 = y0;

		__g.drawString("X = " + mouseX, x0 + _numplayers*w + _MARGIN + 5, y0 + _numplayers*h/3);
		__g.drawString("Y = " + mouseY, x0 + _numplayers*w + _MARGIN + 5, y0 + _numplayers*2*h/3);
		for(int j=0;j < _numplayers;j++)
		{
			if(cells[mouseX][mouseY].memorylist[j])
			{
				__g.setColor(_players[j].color());
				__g.fillRect(x1, y1 + j*h, w, h);
			}
			__g.setColor(_CBLACK);
			__g.drawString(Integer.toString(map[InteractivePlayer][j]), x1 + w/2, y1 + j*h + h/2);
		}
	//	__g.drawString("X = " + mouseX + ", Y = " + mouseY, x0 - 5*_MARGIN, y0 + 2*h);
            } catch (Exception EXC) {
                EXC.printStackTrace();
            }
	} // End - paintComponent

    } // End - ViewPanel Class
    
    //********************************************
    //*
    //* Control Panel
    //*
    //********************************************
    private final class ControlPanel extends JPanel implements ActionListener, ItemListener, Serializable {
        JTabbedPane  _tab;
        JPanel       _conf;
        JPanel       _info;
        final        int         _CWIDTH = 300;
        final        int         _CHEIGHT = 350;
        final        int         _CPANEL_WIDTH = _CWIDTH;
        final        int         _CPANEL_HEIGHT = 21;
        final        int         _CPLAYER_NAME_LENGTH = 20;
        final        ImageIcon   _CSTEP_ICON  = new ImageIcon("Images/marble_step.gif");
        final        ImageIcon   _CSTOP_ICON  = new ImageIcon("Images/marble_stop.gif");
//        final        ImageIcon   _CGRAPH_ICON  = new ImageIcon("Images/marble_graph.gif");
        final        ImageIcon   _CPLAY_ICON  = new ImageIcon("Images/marble_play.gif");
        final        ImageIcon   _CRESET_ICON = new ImageIcon("Images/marble_reset.gif");
        final        Color       _CDISABLED_FIELD_COLOR = new Color(1.0f, 1.0f, 1.0f);
        final Font   _CCONTROL_FONT  = new Font("Courier", Font.BOLD, 12);
        final Font   _CCOMBO_FONT = new Font("Courier", Font.BOLD, 10);

        JTextField   _currRoundfield;
        JTextField[] _scores;
        JTextField   _numplayersfield;
        JComboBox[]  _classes;
        JPanel       _infobox;
        JPanel       _confbox;
        JButton      _play;
        JButton      _step;
        JButton      _reset;        
	JButton      _stop;
//	JButton      _graph;

        JTextField   _maxroundsfield;
        JTextField   _Lfield;
        JTextField   _Wfield;
        JTextField[] _playercopiesfield;
        JTextField   _resetfield;

        NumberFormat _nf;

        //********************************************
        //*
        //* Constructor
        //*
        //********************************************
        public ControlPanel() throws Exception {
            super();

            SlotPanel       slot;
            JPanel          box;
            JLabel          label;
            int             _MAX;
            StringBuffer    SB;
            String          name;

            setLayout(new BorderLayout());
            setBorder(BorderFactory.createCompoundBorder(BorderFactory.createRaisedBevelBorder(),
                                                         BorderFactory.createLoweredBevelBorder()));
            setPreferredSize(new Dimension(_CWIDTH, _CHEIGHT));
            setMinimumSize(new Dimension(_CWIDTH, _CHEIGHT));
            setFont(_CCONTROL_FONT);

            _info = new JPanel();
            _info.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createRaisedBevelBorder(),
                                                         BorderFactory.createLoweredBevelBorder()));
            _info.setLayout(new BorderLayout());
            _info.setPreferredSize(new Dimension(_CWIDTH, _CHEIGHT));
            _info.setMinimumSize(new Dimension(_CWIDTH, _CHEIGHT));
            _info.setFont(_CCONTROL_FONT);

            _reset = new JButton(_CRESET_ICON);
            _reset.addActionListener(this);
            _step = new JButton(_CSTEP_ICON);
            _step.addActionListener(this);
            _play = new JButton(_CPLAY_ICON);
            _play.addActionListener(this);

	    /* amg2006 adding stop button */
	    _stop = new JButton(_CSTOP_ICON);
	    _stop.addActionListener(this);

//	    _graph = new JButton(_CGRAPH_ICON);
//	    _graph.addActionListener(this);

            box = new JPanel();
            box.setLayout(new BoxLayout(box, BoxLayout.Y_AXIS));
            slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
            _input = new JTextField();
            _input.setFont(_CCONTROL_FONT);
            _input.addActionListener(this);
            _input.setText(Integer.toString(InteractivePlayer));
            label = new JLabel("Interactive Player: ");
            label.setFont(_CCONTROL_FONT);
            slot.add(label, _input);
            box.add(slot);

            slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
            _currRoundfield = new JTextField();
            _currRoundfield.setEditable(false);
            _currRoundfield.setFont(_CCONTROL_FONT);
            label = new JLabel("Round No.:          ");
            label.setFont(_CCONTROL_FONT);
            slot.add(label, _currRoundfield);
            box.add(slot);

            _MAX = numPlayers();
            _scores = new JTextField[_MAX];
            for (int i=0; i < _MAX; i++) {
                _scores[i] = new JTextField();
                _scores[i].setEditable(false);
                _scores[i].setFont(_CCONTROL_FONT);
                slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
                label = new JLabel(Util.adjustString(_players[i].name(), _CPLAYER_NAME_LENGTH));
                label.setForeground(_players[i].color());
                label.setFont(_CCONTROL_FONT);
                slot.add(label, _scores[i]);
                box.add(slot);
            }
	    
            _info.add(box, BorderLayout.CENTER);
            _infobox = box;

            _conf = new JPanel();
            _conf.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createRaisedBevelBorder(),
                                                         BorderFactory.createLoweredBevelBorder()));
            _conf.setLayout(new BorderLayout());
            _conf.setPreferredSize(new Dimension(_CWIDTH, _CHEIGHT));
            _conf.setMinimumSize(new Dimension(_CWIDTH, _CHEIGHT));
            _conf.setFont(_CCONTROL_FONT);
            
            box = new JPanel();
            box.setLayout(new BoxLayout(box, BoxLayout.Y_AXIS));            

            slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
            _maxroundsfield = new JTextField();
            _maxroundsfield.setFont(_CCONTROL_FONT);
            _maxroundsfield.setText(Integer.toString(_config.maxRounds()));
            _maxroundsfield.addActionListener(this);
            label = new JLabel("maxRounds:      ");
            label.setFont(_CCONTROL_FONT);
            slot.add(label, _maxroundsfield);
            box.add(slot);
            
            slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
            _Lfield = new JTextField();
            _Lfield.setFont(_CCONTROL_FONT);
            _Lfield.setText(Integer.toString(_config.GridL()));
            _Lfield.addActionListener(this);
            label = new JLabel("Grid L:         ");
            label.setFont(_CCONTROL_FONT);
            slot.add(label, _Lfield);
            box.add(slot);
            
            slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
            _Wfield = new JTextField();
            _Wfield.setFont(_CCONTROL_FONT);
            _Wfield.setText(Integer.toString(_config.GridW()));
            _Wfield.addActionListener(this);
            label = new JLabel("Grid W:         ");
            label.setFont(_CCONTROL_FONT);
            slot.add(label, _Wfield);
            box.add(slot);
            
	    /*
            slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
            _numplayersfield = new JTextField();
            _numplayersfield.setFont(_CCONTROL_FONT);
            _numplayersfield.setText(Integer.toString(_MAX));
            _numplayersfield.addActionListener(this);
            label = new JLabel("NumPlayers:     ");
            label.setFont(_CCONTROL_FONT);
            slot.add(label, _numplayersfield);
            box.add(slot);
	    */

	    _playercopiesfield = new JTextField[_classlist.length];
	    String spaces = "                                    ";
	    for(int i=0;i < _classlist.length;i++)
	    {
            	slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
            	_playercopiesfield[i] = new JTextField();
            	_playercopiesfield[i].setFont(_CCONTROL_FONT);
            	_playercopiesfield[i].setText(Integer.toString(_config.getPlayerCopies(i)));
            	_playercopiesfield[i].addActionListener(this);
		String str = _classlist[i].toString();
		str = str.substring(str.indexOf('.') + 1);
            	label = new JLabel(str + spaces.substring(1, 20 - str.length()));
            	label.setFont(_CCONTROL_FONT);
            	slot.add(label, _playercopiesfield[i]);
            	box.add(slot);
	    }
            
	    
	    /*
            _classes = new JComboBox[_MAX];
            for (int i=0; i < _MAX; i++) {               
                _classes[i] = new JComboBox(_classlist);
                _classes[i].setSelectedItem(_players[i].playerClass());
                _classes[i].addItemListener(this);
                _classes[i].setFont(_CCOMBO_FONT);
                slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
                label = new JLabel("["+i+"]:  ");
                label.setFont(_CCONTROL_FONT);
                slot.add(label, _classes[i]);
                box.add(slot);                                    
            }                        
	    */

            slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
            box.add(slot);

            slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
            _resetfield = new JTextField();
            _resetfield.setFont(_CCONTROL_FONT);
//            _resetfield.setText(_config.resetvalue());
            _resetfield.setText("0");
            _resetfield.addActionListener(this);
            label = new JLabel("Reset All:         ");
            label.setFont(_CCONTROL_FONT);
            slot.add(label, _resetfield);
            box.add(slot);


            _conf.add(box, BorderLayout.CENTER);
            _confbox = box;

            _tab = new JTabbedPane();
            _tab.add("Information", _info);                  
            _tab.add("Configuration", _conf);
            add(_tab, BorderLayout.CENTER);            
            
            _nf = NumberFormat.getInstance();
            _nf.setMinimumFractionDigits(2);
            _nf.setMaximumFractionDigits(2);

        }        

        //********************************************
        //*
        //* ActionListener Interface
        //*
        //********************************************
        public void actionPerformed(ActionEvent __event) {
            Object source = __event.getSource();
            ParseValue pv = null;
            JComboBox[] tmp;
            Class[] tmpcls;
            char[] moves;
            int prev;
            int curr;
            int _MAX;
            SlotPanel slot;
            JLabel label;
            double[] scores;

            try {
                if (source == _step) {
                    step();
                    this.refresh();
                    Friends.this.refresh();
                    return;
                }
                if (source == _play) {
		    new StopListener(this).start();
                    return;
                }
                if (source == _reset) {
                    reset();
                    return;
                }
		if (source == _stop) {
		    _state = _CFINISHED;
		}

                if (source == _input) {
                    if (_state == _CFINISHED) {
                        return;
                    }
		    pv = ParseValue.parseIntegerValue(_input.getText(), 0, _numplayers - 1);
                    if (pv.isValid()) {
                        InteractivePlayer = ((Integer) pv.value()).intValue();
                    this.refresh();
                    Friends.this.refresh();
                    } else {
                        println("Invalid Input");
                    }
                    return;
                }
                if (source == _maxroundsfield) {
                    pv = ParseValue.parseIntegerValue(_maxroundsfield.getText(), _CMIN_ROUNDS, _CMAX_ROUNDS);
                    if (pv.isValid()) {
                        _config.setMaxRounds(((Integer) pv.value()).intValue());
                        _ui.configure(_config);
                    } else {
                        println("Invalid Input");
                    }
                }
                if (source == _Lfield) {
                    pv = ParseValue.parseIntegerValue(_Lfield.getText(), _CMIN_L, _CMAX_L);
                    if (pv.isValid()) {
                        _config.setGridL(((Integer) pv.value()).intValue());
                        _ui.configure(_config);
                    } else {
                        println("Invalid Input");
                    }
                }
                if (source == _Wfield) {
                    pv = ParseValue.parseIntegerValue(_Wfield.getText(), _CMIN_W, _CMAX_W);
                    if (pv.isValid()) {
                        _config.setGridW(((Integer) pv.value()).intValue());
                        _ui.configure(_config);
                    } else {
                        println("Invalid Input");
                    }
                }
                if (source == _resetfield) {
                    pv = ParseValue.parseIntegerValue(_resetfield.getText(), 0, 1000);
                    if (pv.isValid()) {
			    for(int i=0;i < _classlist.length;i++)
			    {
                    		    _config.setPlayerCopies(i, ((Integer) pv.value()).intValue());
			    }
                        _ui.configure(_config);
                    } else {
                        println("Invalid Input");
                    }
                }
		for(int i=0;i < _classlist.length;i++)
		{
			if(source == _playercopiesfield[i])
			{
                    		pv = ParseValue.parseIntegerValue(_playercopiesfield[i].getText(), 0, 1000);
                    		if (pv.isValid()) {
                    		    _config.setPlayerCopies(i, ((Integer) pv.value()).intValue());
                    		    _ui.configure(_config);
                    		} else {
                    		    println("Invalid Input");
                    		}
			}
		}
                if (source == _numplayersfield) {
                    pv = ParseValue.parseIntegerValue(_numplayersfield.getText(), _CMIN_PLAYERS, _CMAX_PLAYERS);
                    if (pv.isValid()) {
                        prev = _config.numPlayers();
                        curr = ((Integer) pv.value()).intValue();
                        if (prev == curr) {
                            return;
                        }
                        if (curr > prev) {
                            tmp = _classes;
                            _classes = new JComboBox[curr];
                            System.arraycopy(tmp, 0, _classes, 0, prev);
                            tmpcls = new Class[curr];
                            System.arraycopy(_config.playerList(), 0, tmpcls, 0, prev);
                            for (int i=prev; i < curr; i++) {
                                _classes[i] = new JComboBox(_classlist);
                                _classes[i].addItemListener(this);
                                _classes[i].setFont(_CCOMBO_FONT);
                                slot = new SlotPanel(_CPANEL_WIDTH, _CPANEL_HEIGHT);
                                label = new JLabel("["+i+"]:  ");
                                label.setFont(_CCONTROL_FONT);
                                slot.add(label, _classes[i]);
                                _confbox.add(slot);
                                tmpcls[i] = (Class) _classes[i].getSelectedItem();
                            }
                            _config.setPlayerList(tmpcls);
                        }
                        if (curr < prev) {
                            tmp = new JComboBox[curr];
                            System.arraycopy(_classes, 0, tmp, 0, curr);
                            tmpcls = new Class[curr];
                            System.arraycopy(_config.playerList(), 0, tmpcls, 0,  curr);
                            for (int i=curr; i < prev; i++) {
                                _confbox.remove(_confbox.getComponents().length - 1);
                            }
                            _classes = tmp;
                            _config.setPlayerList(tmpcls);
                        }
                        _ui.configure(_config);
                        repaint();
                        Friends.this.refresh();
                    } else {
                        println("Invalid Input");
                    }
                }                                                                                
            } catch (Exception EXC) {
                System.out.println(EXC.getMessage());
                EXC.printStackTrace();
            }
        }        
        
        //********************************************
        //*
        //* ItemListener Interface
        //*
        //********************************************            
        public void itemStateChanged(ItemEvent __event) {
            Object source = __event.getSource();
            int _MAX = _classes.length;
            
            try {
                for (int i=0; i < _MAX; i++) {
                    if (source == _classes[i]) {
                        _config.setPlayer(i, (Class) _classes[i].getSelectedItem());
                        _ui.configure(_config);
                    }                                
                }                        
            } catch (Exception EXC) {
                System.out.println(EXC.getMessage());
                EXC.printStackTrace();
            }
        }
         
        //********************************************
        //*
        //* Score Updater
        //*
        //********************************************
        public void refresh() throws Exception {
            int _MAX = numPlayers();

            for (int i=0; i < _MAX; i++) {
//                _scores[i].setText(Integer.toString(pop));
                //_scores[i].setText(_players[i].result());
            }
            _currRoundfield.setText(Integer.toString(currRound()));
        }                            

        
        //********************************************
        //*
        //* Action Tool Exporter
        //*
        //********************************************
        public JButton[] exportTools() {
	    /* amg2006 changed to 4 */
	    // djv changed to 5
            JButton[] ret = new JButton[4];
            ret[0] = _reset;
            ret[1] = _step;
            ret[2] = _play;
	    ret[3] = _stop;
//	    ret[4] = _graph;
            return ret;
        }
    }

    class StopListener extends Thread
    {
	private ControlPanel controlPanel;	
	public StopListener(ControlPanel cp) { 
	    controlPanel=cp;
	} 
	public void run ()
	{
		try
		{
		    while (GUI._amoeba.step()) {
			GUI._amoeba.refresh();
			controlPanel.refresh();
		    }		
			GUI._amoeba.refresh();
			controlPanel.refresh();
		}
		catch (Exception e)
		{
		    System.out.println("unexpected exception caught in run");
	                e.printStackTrace();
		}
	}
    }

} // end - class Friends


