//***********************************************************
//*
//* File:           PlayerWrapper.java
//* Author:         Abhinav Kamra
//* Contact:        kamra@cs.columbia.edu
//* Update:         10.22.2004
//*
//* Description:    Compositional wrapper for IFCPlayer
//*                 objects.
//*
//***********************************************************

package Friends;

import java.util.*;
import java.io.*;
import java.awt.Color;

public final class PlayerWrapper implements Serializable, IFCPlayer {

	double		_score;
	IFCPlayer	_player;
	Class		_class;
	boolean[]	_infolist;
	int		xloc;
	int		yloc;

	PlayerWrapper(Class __class) throws Exception {
        _class = __class;
	}
    
    private void register_priv(Friends __amoeba) throws Exception {
        try {
            _player.register(__amoeba);
	    _infolist = new boolean[__amoeba.numPlayers()];
	    _score = 0.0;
        } catch (Exception EXC) {
            System.out.println(EXC.getMessage());
            EXC.printStackTrace();
	    System.out.println("Player " + _class + " threw an Exception in register()");
        }
    }


	public void register(Friends __amoeba) throws Exception {
        _player = (IFCPlayer) _class.newInstance();
        register_priv(__amoeba);

    }

    public void register(Friends __amoeba, IFCPlayer __player) throws Exception {
        _player = __player;
        _class = _player.getClass();
        register_priv(__amoeba);
    }

	public String name() throws Exception {
        try {
	        return _player.name();
        } catch (Exception EXC) {
            System.out.println(EXC.getMessage());
            EXC.printStackTrace();
	    System.out.println("Player " + _class + " threw an Exception in name()");
	    return "Anonymous";
        }
    }

    public boolean[] GetInfo() throws Exception {
	    boolean[] RET = new boolean[_infolist.length];
	    System.arraycopy(_infolist, 0, RET, 0, _infolist.length);
	    return RET;
    }
    
    public int SetInfo(int i) throws Exception {
	    _infolist[i] = true;
	    return 1;
    }
    
    public int SetInfo(boolean[] allinfo) throws Exception {
	    int RET = 0;
	    for(int i=0;i < allinfo.length;i++)
	    {
		    _infolist[i] |= allinfo[i];
		    RET += (_infolist[i])?(1):(0);
	    }
	    return RET;
    }

    public Color color() throws Exception {
        try {
            return _player.color();
        } catch (Exception EXC) {
            System.out.println(EXC.getMessage());
            EXC.printStackTrace();
	    System.out.println("Player " + _class + " threw an Exception in color()");
	    return new Color(1.0f, 1.0f, 0.9f);
        }
    }

    public Class playerClass() throws Exception {
        return _class;
    }

    public void gameOver() throws Exception {
        try {
            if (_player instanceof IFCPersistentPlayer) {
                ((IFCPersistentPlayer) _player).gameOver();
            }
        } catch (Exception EXC) {
            System.out.println(EXC.getMessage());
            EXC.printStackTrace();
	    System.out.println("Player " + _class + " threw an Exception in gameOver()");
        }
    }

    public int move(Move input) throws Exception {
        try {
            int RET = _player.move(input);
            
            return RET;

        } catch (Exception EXC) {
            System.out.println(EXC.getMessage());
            EXC.printStackTrace();
	    System.out.println("Player " + _class + " threw an Exception in move()");
            return _CSTAYPUT;
        }
    }

    public boolean interactive() throws Exception {
        try {
            return _player.interactive();
        } catch (Exception EXC) {
            System.out.println(EXC.getMessage());
            EXC.printStackTrace();
	    System.out.println("Player " + _class + " threw an Exception in interactive()");
	    return false;
        }
    }

	public double score() {
	    return _score;
    }

    public int x() {
        return xloc;
    }

    public int y() {
        return yloc;
    }

    public void setLocation(int x, int y) {
        xloc = x;
        yloc = y;
    }

    public void setScore(double __score) {
        _score = __score;
    }

    public IFCPlayer player() {
        return _player;
    }

    public String toString() {
	    try {
    		StringBuffer SB = new StringBuffer("[");

            SB.append(_player.name());
		    SB.append("]");
    		return new String(SB);
        } catch (Exception EXC) {
            return EXC.getMessage();
        }
	}
}
